import wandb

from lm_buddy.configs.jobs import (
    EvaluationJobConfig,
    FinetuningJobConfig,
    HuggingFaceEvalJobConfig,
    JobConfig,
    LMHarnessJobConfig,
    PrometheusJobConfig,
    RagasJobConfig,
)
from lm_buddy.jobs.common import EvaluationResult, FinetuningResult, JobType
from lm_buddy.jobs.evaluation.hf_evaluate import run_hf_evaluation
from lm_buddy.jobs.evaluation.lm_harness import run_lm_harness
from lm_buddy.jobs.evaluation.prometheus import run_prometheus
from lm_buddy.jobs.evaluation.ragas import run_ragas
from lm_buddy.jobs.finetuning import run_finetuning
from lm_buddy.paths import strip_path_prefix
from lm_buddy.tracking.run_utils import WandbResumeMode


class LMBuddy:
    """Your buddy in the (L)LM space.

    Simple wrapper around executable functions for tasks available in the library.
    """

    # TODO: Store some configuration (e.g., tracking info, name) globally on the buddy
    def __init__(self):
        pass

    def _generate_artifact_lineage(
        self, config: JobConfig, results: list[wandb.Artifact], job_type: JobType
    ) -> None:
        """Link input artifacts and log output artifacts to a run.

        A no-op if no tracking config is available.
        """
        if config.tracking is not None:
            with wandb.init(
                name=config.name,
                job_type=job_type,
                resume=WandbResumeMode.ALLOW,
                **config.tracking.model_dump(),
            ) as run:
                for path in config.artifact_paths():
                    artifact_name = strip_path_prefix(path)
                    run.use_artifact(artifact_name)
                for artifact in results:
                    artifact = run.log_artifact(artifact)
                    artifact.wait()

    def finetune(self, config: FinetuningJobConfig) -> FinetuningResult:
        """Run a supervised finetuning job with the provided configuration."""
        result = run_finetuning(config)
        self._generate_artifact_lineage(config, result.artifacts, JobType.FINETUNING)
        return result

    def evaluate(self, config: EvaluationJobConfig) -> EvaluationResult:
        """Run an evaluation job with the provided configuration.

        The underlying evaluation framework is determined by the configuration type.
        """
        match config:
            case LMHarnessJobConfig() as lm_harness_config:
                result = run_lm_harness(lm_harness_config)
            case PrometheusJobConfig() as prometheus_config:
                result = run_prometheus(prometheus_config)
            case RagasJobConfig() as ragas_config:
                result = run_ragas(ragas_config)
            case HuggingFaceEvalJobConfig() as hf_eval_config:
                result = run_hf_evaluation(hf_eval_config)
            case _:
                raise ValueError(f"Invlid configuration for evaluation: {type(config)}")
        self._generate_artifact_lineage(config, result.artifacts, JobType.EVALUATION)
        return result



import string
from typing import Any

from datasets import Dataset


def format_dataset_with_prompt(
    dataset: Dataset,
    prompt_template: str,
    output_field: str,
    *,
    num_proc: int = 1,
) -> Dataset:
    fields = _get_template_fields(prompt_template)

    missing_columns = fields.difference(dataset.column_names)
    if missing_columns:
        raise ValueError(
            f"Dataset is missing columns for the following prompt fields: {missing_columns}"
        )

    # Formatter is typed for single examples
    def formatting_func(examples: dict[str, Any]):
        format_data = {k: examples[k] for k in fields}
        examples[output_field] = prompt_template.format(**format_data)
        return examples

    return dataset.map(formatting_func, num_proc=num_proc, batched=False)


def _get_template_fields(template: str) -> set[str]:
    """Returns the fields in the template."""
    parsed = [x for x in string.Formatter().parse(template) if x[1] is not None]
    fields = {field for _, field, _, _ in parsed}
    return fields



from enum import Enum
from typing import Any

import wandb
from wandb.apis.public import Run as ApiRun

from lm_buddy.configs.wandb import WandbRunConfig


class WandbResumeMode(str, Enum):
    """Enumeration of modes for resuming a W&B run.

    This is not an exahustive list of the values that can be passed to the W&B SDK
    (Docs: https://docs.wandb.ai/ref/python/init), but just those commonly used within the package.
    """

    ALLOW = "allow"
    MUST = "must"
    NEVER = "never"


def get_run_from_api(config: WandbRunConfig) -> ApiRun:
    """Retrieve a run from the W&B API."""
    api = wandb.Api()
    return api.run(config.wandb_path())


def get_run_summary(config: WandbRunConfig) -> dict[str, Any]:
    """Get the summary dictionary attached to a W&B run."""
    run = get_run_from_api(config)
    return dict(run.summary)


def update_wandb_summary(config: WandbRunConfig, metrics: dict[str, Any]) -> None:
    """Update a run's summary with the provided metrics."""
    run = get_run_from_api(config)
    run.summary.update(metrics)
    run.update()



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any

import pandas as pd
import wandb


class JobType(str, Enum):
    """Enumeration of logical job types runnable via the LM Buddy."""

    PREPROCESSING = "preprocessing"
    FINETUNING = "finetuning"
    EVALUATION = "evaluation"


@dataclass
class JobResult:
    artifacts: list[wandb.Artifact]


@dataclass
class FinetuningResult(JobResult):
    """Result from a finetuning task."""

    checkpoint_path: Path | None
    metrics: dict[str, Any] | None
    is_adapter: bool


@dataclass
class EvaluationResult(JobResult):
    """Result from an evaluation task, containing aggregate metrics and artifacts."""

    tables: dict[str, pd.DataFrame]
    dataset_path: Path | None



from pathlib import Path
from typing import TypeVar

from lm_buddy.configs.jobs.common import JobConfig

ConfigType = TypeVar("ConfigType", bound=JobConfig)


def parse_config_option(config_cls: type[ConfigType], config: str) -> ConfigType:
    """Parse the config option string from the CLI.

    If it corresponds to a path that exists, attempt to load the config from YAML file.
    If not, attempt to parse it as a JSON string.
    """
    if Path(config).exists():
        return config_cls.from_yaml_file(config)
    else:
        return config_cls.model_validate_json(config)



from pydantic import BaseModel


def copy_pydantic_json(model: BaseModel) -> BaseModel:
    """Copy a Pydantic model through round-trip JSON serialization."""
    return model.__class__.model_validate_json(model.model_dump_json())



import os
from pathlib import Path

LM_BUDDY_HOME_PATH: str = os.getenv(
    "LM_BUDDY_HOME",
    str(Path.home() / ".lm_buddy"),
)

LM_BUDDY_RESULTS_PATH: str = os.getenv(
    "LM_BUDDY_RESULTS",
    f"{LM_BUDDY_HOME_PATH}/results",
)



import click

from lm_buddy.cli import evaluate, finetune


@click.group(name="LM Buddy CLI", help="Entrypoints for the LM Buddy.")
def cli():
    pass


cli.add_command(finetune.command)
cli.add_command(evaluate.group)



from pydantic import Field, conlist, field_validator, model_validator

from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.configs.huggingface import (
    AutoModelConfig,
    AutoTokenizerConfig,
    DatasetConfig,
    QuantizationConfig,
)
from lm_buddy.configs.jobs.common import JobConfig
from lm_buddy.configs.vllm import VLLMCompletionsConfig
from lm_buddy.paths import AssetPath


class HuggingFaceEvaluationConfig(LMBuddyConfig):
    """Misc settings provided to an lm-harness evaluation job."""

    metrics: conlist(str, min_length=1)
    use_pipeline: bool = False
    enable_tqdm: bool = False
    max_samples: int | None = None
    storage_path: str | None = None
    return_input_data: bool = False
    return_predictions: bool = False


class HuggingFaceEvalJobConfig(JobConfig):
    """Configuration to run a HuggingFace evaluation job."""

    dataset: DatasetConfig = Field(
        description="Dataset of text completions to evaluate using the Prometheus judge model."
    )
    evaluation: HuggingFaceEvaluationConfig
    model: AutoModelConfig | VLLMCompletionsConfig
    quantization: QuantizationConfig | None = None
    tokenizer: AutoTokenizerConfig

    @model_validator(mode="before")
    def ensure_tokenizer_config(cls, values):
        """Set the tokenizer to the model path when not explicitly provided."""
        if values.get("tokenizer") is None:
            values["tokenizer"] = {}
            match values["model"]:
                case str() as model_path:
                    values["tokenizer"]["path"] = model_path
                case dict() as model_data:
                    # if dict we might have model.path specified
                    # if we don't it is VLLMCompletion and we are ok
                    # with anything as it will be ignored
                    if model_data.get("path") is None:
                        values["tokenizer"]["path"] = "oai://tokenizer"
                    else:
                        values["tokenizer"]["path"] = model_data.get("path")
                case AutoModelConfig() as model_config:
                    values["tokenizer"]["path"] = model_config.path
                # No fallback necessary, downstream validation will flag invalid model types
        return values

    @field_validator("model", mode="before")
    def validate_model_arg(cls, x):
        """Allow for passing just a path string as the model argument."""
        if isinstance(x, str):
            return AutoModelConfig(path=x)
        return x

    @field_validator("tokenizer", mode="before")
    def validate_tokenizer_arg(cls, x):
        """Allow for passing just a path string as the tokenizer argument."""
        if isinstance(x, str):
            return AutoTokenizerConfig(path=x)
        return x

    def asset_paths(self) -> list[AssetPath]:
        match self.model:
            case AutoModelConfig() as config:
                return {
                    self.dataset.path,
                    self.evaluation.output_path,
                    config.path,
                    self.tokenizer.path,
                }
            case VLLMCompletionsConfig() as config if config.inference.engine is not None:
                return {self.dataset.path, self.evaluation.output_path, config.inference.engine}
            case _:
                return {}



import re
from enum import Enum
from pathlib import Path
from typing import Annotated

import wandb
from huggingface_hub.utils import HFValidationError, validate_repo_id
from pydantic import AfterValidator
from wandb.sdk.artifacts.exceptions import ArtifactNotLoggedError


class PathPrefix(str, Enum):
    FILE = "file://"
    HUGGINGFACE = "hf://"
    WANDB = "wandb://"
    S3 = "s3://"
    OPENAI = "oai://"


def strip_path_prefix(path: str) -> str:
    """Strip the 'scheme://' prefix from the start of a string."""
    pattern = "^\w+\:\/\/"
    return re.sub(pattern, "", path)


def is_valid_huggingface_repo_id(s: str):
    """
    Simple test to check if an HF model is valid using HuggingFace's tools.
    Sadly, theirs throws an exception and has no return.

    Args:
        s: string to test.
    """
    try:
        validate_repo_id(s)
        return True
    except HFValidationError:
        return False


def validate_asset_path(path: str) -> "AssetPath":
    raw_path = strip_path_prefix(path)
    if path.startswith(PathPrefix.FILE):
        if not Path(raw_path).is_absolute():
            raise ValueError(f"'{raw_path}' is not an absolute file path.")
    elif path.startswith(PathPrefix.HUGGINGFACE):
        if not is_valid_huggingface_repo_id(raw_path):
            raise ValueError(f"'{raw_path}' is not a valid HuggingFace repo ID.")
    elif path.startswith(PathPrefix.WANDB):
        # TODO: Validate the W&B path structure?
        pass
    elif path.startswith(PathPrefix.S3):
        # TODO: Validate the S3 path structure?
        # e.g. if the assumption is we always want a file or a dir, we could
        # use https://s3pathlib.readthedocs.io to verify (.is_file() or ._is_dir())
        pass
    elif path.startswith(PathPrefix.OPENAI):
        # TODO: Validate the OAI path structure?
        pass
    else:
        allowed_prefixes = {x.value for x in PathPrefix}
        raise ValueError(f"'{path}' does not begin with an allowed prefix: {allowed_prefixes}")
    return path


AssetPath = Annotated[str, AfterValidator(lambda x: validate_asset_path(x))]


def format_file_path(path: str | Path) -> AssetPath:
    path = Path(path).absolute()
    return f"{PathPrefix.FILE.value}{path}"


def format_huggingface_path(repo_id: str) -> AssetPath:
    return f"{PathPrefix.HUGGINGFACE.value}{repo_id}"


def format_artifact_path(artifact: wandb.Artifact) -> AssetPath:
    try:
        return f"{PathPrefix.WANDB.value}{artifact.qualified_name}"
    except ArtifactNotLoggedError as e:
        msg = (
            "Unable to construct an `AssetPath` from artifact missing project/entity fields. "
            "Make sure to log the artifact before calling this method."
        )
        raise ValueError(msg) from e



import click

from lm_buddy import LMBuddy
from lm_buddy.cli.utils import parse_config_option
from lm_buddy.configs.jobs import FinetuningJobConfig


@click.command(name="finetune", help="Run an LM Buddy finetuning job.")
@click.option("--config", type=str)
def command(config: str) -> None:
    config = parse_config_option(FinetuningJobConfig, config)
    buddy = LMBuddy()
    buddy.finetune(config)



import evaluate
import numpy as np
from loguru import logger


class EvaluationMetrics:
    def __init__(self, metrics):
        self._supported_metrics = {
            "rouge": self._rouge,
            "meteor": self._meteor,
            "bertscore": self._bertscore,
        }

        # chosen metrics are the intersection between the provided and the supporterd ones
        self._chosen_metrics = set(metrics) & set(self._supported_metrics.keys())
        # unsupported metrics are the difference between the provided and the supporterd ones
        self._unsupported_metrics = set(metrics) - set(self._supported_metrics.keys())

        if len(self._chosen_metrics) == 0:
            logger.info("No valid metrics selected")
        else:
            logger.info(f"Chosen metrics: {self._chosen_metrics}")

        if len(self._unsupported_metrics) > 0:
            logger.info(f"Unsupported metrics: {self._unsupported_metrics}")

    def _rouge(self, pred, ref):
        ev = evaluate.load("rouge")

        # compute with use_aggregator = False to get individual scores
        evals = ev.compute(predictions=pred, references=ref, use_aggregator=False)

        # calculate mean for each of the submetrics (rouge1, rouge2, rougeL, rougeLsum)
        for k in ["rouge1", "rouge2", "rougeL", "rougeLsum"]:
            evals[f"{k}_mean"] = np.mean(evals[k])

        return evals

    def _meteor(self, pred, ref):
        ev = evaluate.load("meteor")

        # initialize dictionary with metric name
        evals = {"meteor": []}

        # run sample-wise evals (as default implementation only returns mean value)
        for p, r in zip(pred, ref):
            evals["meteor"].append(ev.compute(predictions=[p], references=[r])["meteor"])

        # calculate mean
        evals["meteor_mean"] = np.mean(evals["meteor"])

        return evals

    def _bertscore(self, pred, ref):
        ev = evaluate.load("bertscore")

        # calculate evals (the default is not to aggregate them)
        evals = ev.compute(predictions=pred, references=ref, lang="en")

        # calculate mean for each of the submetrics (precision, recall, f1)
        for k in ["precision", "recall", "f1"]:
            evals[f"{k}_mean"] = np.mean(evals[k])

        return evals

    def run_all(self, pred, ref):
        results = {}

        for metric in self._chosen_metrics:
            results[metric] = self._supported_metrics[metric](pred, ref)

        return results



from pathlib import Path

from datasets import Dataset
from langchain_community.embeddings.huggingface import HuggingFaceEmbeddings
from langchain_openai import ChatOpenAI
from loguru import logger
from ragas import evaluate as ragas_evaluate
from ragas.metrics import answer_relevancy, context_precision, context_recall, faithfulness

from lm_buddy.configs.jobs.ragas import RagasJobConfig
from lm_buddy.constants import LM_BUDDY_RESULTS_PATH
from lm_buddy.jobs.asset_loader import HuggingFaceDatasetLoader, HuggingFaceModelLoader
from lm_buddy.jobs.common import EvaluationResult
from lm_buddy.preprocessing import format_dataset_with_prompt
from lm_buddy.tracking.artifact_utils import (
    ArtifactType,
    build_directory_artifact,
    default_artifact_name,
)

RAGAS_METRICS_MAP = {
    "faithfulness": faithfulness,
    "answer_relevancy": answer_relevancy,
    "context_recall": context_recall,
    "context_precision": context_precision,
}


def run_eval(config: RagasJobConfig) -> Path:
    # load dataset from W&B artifact
    hf_dataset_loader = HuggingFaceDatasetLoader()
    evaluation_dataset = hf_dataset_loader.load_dataset(config.dataset)
    if config.dataset.prompt_template is not None:
        evaluation_dataset = format_dataset_with_prompt(
            evaluation_dataset, config.dataset.prompt_template, config.dataset.text_field
        )

    # Ragas custom model args
    ragas_args = {}

    # Load embedding model
    hf_model_loader = HuggingFaceModelLoader()
    embedding_model = hf_model_loader.resolve_asset_path(config.evaluation.embedding_model)
    ragas_args["embeddings"] = HuggingFaceEmbeddings(model_name=embedding_model)

    # Configure ragas to point to vllm instance for generation
    inference_engine = hf_model_loader.resolve_asset_path(config.judge.inference.engine)
    ragas_args["llm"] = ChatOpenAI(
        model=inference_engine,
        openai_api_key="EMPTY",  # needed to hit custom openai endpoint
        openai_api_base=config.judge.inference.base_url,
        max_tokens=config.judge.max_tokens,
        temperature=config.judge.temperature,
        top_k=config.judge.top_k,
    )

    ragas_metrics = [RAGAS_METRICS_MAP[metric] for metric in config.evaluation.metrics]
    result = ragas_evaluate(dataset=evaluation_dataset, metrics=ragas_metrics, **ragas_args)

    # Return a new dataset with score concatenated
    result_dataset = Dataset.from_pandas(result.to_pandas())

    # Save dataset to disk
    storage_path = config.evaluation.storage_path or LM_BUDDY_RESULTS_PATH
    result_dataset_path = Path(storage_path) / config.name / "ragas"
    result_dataset.save_to_disk(result_dataset_path)

    return result_dataset_path


def run_ragas(config: RagasJobConfig) -> EvaluationResult:
    # Run evaluation
    result_dataset_path = run_eval(config)
    logger.info(f"Ragas evaluation dataset stored at {result_dataset_path}")

    # Create a directory artifact for the HF dataset
    artifact_name = default_artifact_name(config.name, artifact_type=ArtifactType.DATASET)
    dataset_artifact = build_directory_artifact(
        artifact_name=artifact_name,
        artifact_type=ArtifactType.DATASET,
        dir_path=result_dataset_path,
        reference=False,
    )

    return EvaluationResult(
        artifacts=[dataset_artifact],
        dataset_path=result_dataset_path,
        tables={},
    )



from pydantic import Field, field_validator, model_validator

from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.configs.huggingface import (
    AdapterConfig,
    AutoModelConfig,
    AutoTokenizerConfig,
    DatasetConfig,
    QuantizationConfig,
    TrainerConfig,
)
from lm_buddy.configs.jobs.common import JobConfig
from lm_buddy.paths import AssetPath


class FinetuningRayConfig(LMBuddyConfig):
    """Misc settings passed to Ray train for finetuning.

    Includes information for scaling, checkpointing, and runtime storage.
    """

    use_gpu: bool = True
    num_workers: int | None = None
    storage_path: str | None = None  # TODO: This should be set globally somehow


class FinetuningJobConfig(JobConfig):
    """Configuration to submit an LLM finetuning job."""

    model: AutoModelConfig
    dataset: DatasetConfig
    tokenizer: AutoTokenizerConfig
    quantization: QuantizationConfig | None = None
    adapter: AdapterConfig | None = None
    trainer: TrainerConfig = Field(default_factory=TrainerConfig)
    ray: FinetuningRayConfig = Field(default_factory=FinetuningRayConfig)

    @model_validator(mode="before")
    def ensure_tokenizer_config(cls, values):
        """Set the tokenizer to the model path when not explicitly provided."""
        if values.get("tokenizer") is None:
            values["tokenizer"] = {}
            match values["model"]:
                case str() as model_path:
                    values["tokenizer"]["path"] = model_path
                case dict() as model_data:
                    values["tokenizer"]["path"] = model_data["path"]
                case AutoModelConfig() as model_config:
                    values["tokenizer"]["path"] = model_config.path
                # No fallback necessary, downstream validation will flag invalid model types
        return values

    @field_validator("model", mode="before")
    def validate_model_arg(cls, x):
        """Allow for passing just a path string as the model argument."""
        if isinstance(x, str):
            return AutoModelConfig(path=x)
        return x

    @field_validator("tokenizer", mode="before")
    def validate_tokenizer_arg(cls, x):
        """Allow for passing just a path string as the tokenizer argument."""
        if isinstance(x, str):
            return AutoTokenizerConfig(path=x)
        return x

    def asset_paths(self) -> list[AssetPath]:
        return {self.model.path, self.dataset.path, self.tokenizer.path}



from abc import abstractmethod

import torch
from loguru import logger
from openai import OpenAI, OpenAIError
from openai.types import Completion
from transformers import pipeline

from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.configs.huggingface import AutoModelConfig
from lm_buddy.configs.jobs.hf_evaluate import HuggingFaceEvalJobConfig
from lm_buddy.configs.vllm import VLLMCompletionsConfig
from lm_buddy.jobs.asset_loader import HuggingFaceModelLoader, HuggingFaceTokenizerLoader


class BaseModelClient:
    """
    Abstract class for a model client, used to provide a uniform interface
    (currentnly just a simple predict method) to models served in different
    ways (e.g. HF models loaded locally, OpenAI endpoints, vLLM inference
    servers, llamafile).
    """

    @abstractmethod
    def __init__(self, model: str, config: LMBuddyConfig):
        """
        Used to initialize the model / inference service.
        """
        pass

    @abstractmethod
    def predict(self, prompt: str) -> str:
        """
        Given a prompt, return a prediction.
        """
        pass


class SummarizationPipelineModelClient(BaseModelClient):
    """
    Model client for the huggingface summarization pipeline
    (model is loaded locally).
    """

    def __init__(self, model: str, config: AutoModelConfig):
        self._summarizer = pipeline(
            "summarization",
            model=model,
            device=0 if torch.cuda.is_available() else -1,
        )

    def predict(self, prompt):
        # summarizer output is a list (1 element in this case) of dict with key = "summary_text"
        # TODO: bring summarizer parameters out at some point (not needed at the moment)
        pred = self._summarizer(prompt, min_length=30, do_sample=False)
        return pred[0]["summary_text"]


class HuggingFaceModelClient(BaseModelClient):
    """
    Model client for HF models (model is loaded locally, both Seq2SeqLM
    and CausalLM are supported).
    - Provide model path to load the model locally
    - Make sure you add quantization details if the model is too large
    - Optionally, add a tokenizer (the one matching the specified model name is the default)
    """

    def __init__(self, model: str, config: HuggingFaceEvalJobConfig):
        self._config = config
        self._device = "cuda" if torch.cuda.is_available() else "cpu"

        hf_model_loader = HuggingFaceModelLoader()
        hf_tokenizer_loader = HuggingFaceTokenizerLoader()
        self._model = hf_model_loader.load_pretrained_model(config.model).to(self._device)
        self._tokenizer = hf_tokenizer_loader.load_pretrained_tokenizer(config.tokenizer)

    def predict(self, prompt):
        inputs = self._tokenizer(prompt, truncation=True, padding=True, return_tensors="pt").to(
            self._device
        )
        generated_ids = self._model.generate(**inputs, max_new_tokens=256)
        return self._tokenizer.batch_decode(generated_ids, skip_special_tokens=True)[0]


class OpenAIModelClient(BaseModelClient):
    """
    Model client for models served via openai-compatible API.
    For OpenAI models:
    - The base_url is fixed
    - Choose an engine name (see https://platform.openai.com/docs/models)
    - Customize the system prompt if needed

    For compatible models:
    - Works with local/remote vLLM-served models and llamafiles
    - Provide base_url and engine
    - Customize the system prompt if needed
    """

    def __init__(self, model: str, config: VLLMCompletionsConfig):
        self._config = config

        hf_model_loader = HuggingFaceModelLoader()
        self._engine = hf_model_loader.resolve_asset_path(config.inference.engine)
        self._system = config.inference.system_prompt
        self._client = OpenAI(base_url=model)

    def _openai_chat_completion(
        self,
        config: VLLMCompletionsConfig,
        client: OpenAI,
        prompt: str,
        system: str = "You are a helpful assisant.",
    ) -> Completion:
        """Connects to a remote OpenAI-API-compatible endpoint
        and returns a chat completion holding the model's response.
        """

        return self._client.chat.completions.create(
            model=self._engine,
            messages=[{"role": "system", "content": system}, {"role": "user", "content": prompt}],
            max_tokens=config.max_tokens,
            frequency_penalty=config.frequency_penalty,
            temperature=config.temperature,
            top_p=config.top_p,
        )

    def _get_response_with_retries(
        self,
        config: VLLMCompletionsConfig,
        prompt: str,
    ) -> tuple[str, str]:
        current_retry_attempt = 1
        max_retries = 1 if config.inference.max_retries is None else config.inference.max_retries
        while current_retry_attempt <= max_retries:
            try:
                response = self._openai_chat_completion(
                    self._config, self._client, prompt, self._system
                )
                break
            except OpenAIError as e:
                logger.warning(f"{e.message}: Retrying ({current_retry_attempt}/{max_retries})")
                current_retry_attempt += 1
                if current_retry_attempt > max_retries:
                    raise e
        return response

    def predict(self, prompt):
        response = self._get_response_with_retries(self._config, prompt)

        return response.choices[0].message.content



from typing import Literal, get_args

from pydantic import Field, field_validator

from lm_buddy.configs.common import LMBuddyConfig
from lm_buddy.configs.huggingface import AutoModelConfig, DatasetConfig
from lm_buddy.configs.jobs.common import JobConfig
from lm_buddy.configs.vllm import VLLMCompletionsConfig
from lm_buddy.paths import AssetPath

RagasEvaluationMetric = Literal[
    "faithfulness",
    "answer_relevancy",
    "context_recall",
    "context_precision",
]


class RagasEvaluationConfig(LMBuddyConfig):
    """Parameters specifically required for RAGAs Evaluation"""

    metrics: list[RagasEvaluationMetric] = Field(
        default_factory=lambda: list(get_args(RagasEvaluationMetric)),
        description="List of metric names for Ragas evaluation.",
    )
    embedding_model: AssetPath = Field(
        default="hf://sentence-transformers/all-mpnet-base-v2",
        description="Path to embedding model used with the evaluation judge.",
    )
    storage_path: str | None = Field(
        default=None,
        description="Path to store evaluation outputs. Defaults to the `LM_BUDDY_STORAGE` path.",
    )

    @field_validator("embedding_model", mode="before")
    def validate_embedding_model_arg(cls, x):
        """Allow for passing just a path string as the model argument."""
        if isinstance(x, str):
            return AutoModelConfig(path=x)
        return x


class RagasJobConfig(JobConfig):
    """Configuration to run a Ragas evaluation job.

    This job loads a dataset from an existing path on our cluster.
    The dataset must be formatted in the RAG context, with the question, generated answer,
    the contexts (retrieved), and optionally a ground truth field.
    """

    judge: VLLMCompletionsConfig = Field(description="Externally hosted Ragas judge model.")
    dataset: DatasetConfig = Field(
        description="Dataset of text completions to evaluate using the Ragas judge model."
    )
    evaluation: RagasEvaluationConfig = Field(
        default_factory=RagasEvaluationConfig,
        description="Settings for the Ragas evaluation.",
    )

    def asset_paths(self) -> set[AssetPath]:
        paths = {self.dataset.path, self.judge.inference.engine}
        return {x for x in paths if x is not None}



"""
lm-buddy entrypoint to run summary evaluation using huggingface eval
"""

import json
from collections.abc import Iterable
from pathlib import Path

import s3fs
from loguru import logger
from tqdm import tqdm

from lm_buddy.configs.jobs.hf_evaluate import HuggingFaceEvalJobConfig
from lm_buddy.configs.vllm import VLLMCompletionsConfig
from lm_buddy.constants import LM_BUDDY_RESULTS_PATH
from lm_buddy.jobs.asset_loader import (
    HuggingFaceDatasetLoader,
    HuggingFaceModelLoader,
)
from lm_buddy.jobs.common import EvaluationResult
from lm_buddy.jobs.evaluation.metrics import EvaluationMetrics
from lm_buddy.jobs.model_clients import (
    BaseModelClient,
    HuggingFaceModelClient,
    OpenAIModelClient,
    SummarizationPipelineModelClient,
)
from lm_buddy.jobs.utils import timer


@timer
def predict(dataset_iterable: Iterable, model_client: BaseModelClient) -> list:
    predictions = []

    for sample_txt in dataset_iterable:
        predictions.append(model_client.predict(sample_txt))

    return predictions


@timer
def evaluate(predictions: list, ground_truth: list, evaluation_metrics: list):
    em = EvaluationMetrics(evaluation_metrics)
    evaluation_results = em.run_all(predictions, ground_truth)

    return evaluation_results


def save_outputs(config: HuggingFaceEvalJobConfig, evaluation_results: dict) -> Path:
    storage_path = config.evaluation.storage_path

    def save_to_disk(local_path: Path):
        logger.info(f"Storing into {local_path}...")
        local_path.parent.mkdir(exist_ok=True, parents=True)
        with local_path.open("w") as f:
            json.dump(evaluation_results, f)

    def save_to_s3(local_path: Path, storage_path: str):
        s3 = s3fs.S3FileSystem()
        if storage_path.endswith("/"):
            storage_path = "s3://" + str(Path(storage_path[5:]) / config.name / "eval_results.json")
        logger.info(f"Storing into {storage_path}...")
        s3.put_file(local_path, storage_path)

    # generate local temp file ANYWAY
    # (we don't want to lose all eval data if there is an issue wth s3)
    local_path = Path(LM_BUDDY_RESULTS_PATH) / config.name / "eval_results.json"

    try:
        save_to_disk(local_path)

        # copy to s3 and return path
        if storage_path is not None and storage_path.startswith("s3://"):
            save_to_s3(local_path, storage_path)
            return storage_path

        return local_path

    except Exception as e:
        logger.error(e)


def run_eval(config: HuggingFaceEvalJobConfig) -> Path:
    # Init loaders
    hf_dataset_loader = HuggingFaceDatasetLoader()
    hf_model_loader = HuggingFaceModelLoader()

    # Load dataset given its URI
    dataset = hf_dataset_loader.load_dataset(config.dataset)

    # Limit dataset length if max_samples is specified
    if config.evaluation.max_samples is not None:
        dataset = dataset.select(range(config.evaluation.max_samples))

    # Enable / disable tqdm
    input_samples = dataset["examples"]
    dataset_iterable = tqdm(input_samples) if config.evaluation.enable_tqdm else input_samples

    # Choose which model client to use
    if type(config.model) == VLLMCompletionsConfig:
        model_name = config.model.inference.base_url
    else:
        model_name = hf_model_loader.resolve_asset_path(config.model.path)

    if model_name.startswith("http"):
        # run the openai client
        logger.info(f"Using OAI client. Endpoint: {model_name}")
        model_client = OpenAIModelClient(model_name, config.model)
    else:
        # depending on config, use the summarizer pipeline or directly call the model
        # for inference
        if config.evaluation.use_pipeline:
            logger.info(f"Using summarization pipeline. Model: {model_name}")
            model_client = SummarizationPipelineModelClient(model_name, config.model)
        else:
            logger.info(f"Using direct HF model invocation. Model: {model_name}")
            model_client = HuggingFaceModelClient(model_name, config)

    # run inference
    predictions, summarization_time = predict(dataset_iterable, model_client)

    # run evaluation
    ground_truth = dataset["ground_truth"]
    evaluation_results, evaluation_time = evaluate(
        predictions, ground_truth, config.evaluation.metrics
    )

    # add timing to results dict
    evaluation_results["summarization_time"] = summarization_time
    evaluation_results["evaluation_time"] = evaluation_time

    # add input data to results dict
    if config.evaluation.return_input_data:
        evaluation_results["examples"] = dataset["examples"]
        evaluation_results["ground_truth"] = dataset["ground_truth"]

    # add predictions to results dict
    if config.evaluation.return_predictions:
        evaluation_results["predictions"] = predictions

    output_path = save_outputs(config, evaluation_results)
    return output_path


def run_hf_evaluation(config: HuggingFaceEvalJobConfig) -> EvaluationResult:
    # Run eval and store output in local filename
    result_dataset_path = run_eval(config)
    logger.info(f"Summarization eval results stored at {result_dataset_path}")

    return EvaluationResult(
        artifacts=[],
        dataset_path=result_dataset_path,
        tables={},
    )
